unit kuRAFMerge;

//$Revision:1.2$    $RevDate:8/10/2004 3:24:56 PM$
//$NoExpand$

{Copyright 2002 J. Kevin Ugan   All Rights Reserved.}

{See Readme.Txt for license and revision history}

interface

uses
	Windows, Messages, SysUtils, Classes, contnrs, Graphics,
	Controls, Forms, Dialogs;


type
	TRBPage = class(TObject)
	private
		FSourceIndex: Integer;
		FSourcePageNo: Integer;
		FSourceOffset: Integer;
		FPageSize: Integer;
		FDestPageNo: Integer;
		FDestOffset: Integer;
	end;

type
	TPageList = Class(TObjectList);

type
	TkuRAFMerge = class(TComponent)
	private
		{ Private declarations }
		FSourceFiles: TStringList;
		FInputStreams: TObjectList;
		FDestinationFile: TFileName;
		FRenumberPages: Boolean;
		FPages: TPageList;
		FOnProgress: TNotifyEvent;
		FTotBytes: Integer;
		FTotPages: Integer;
		FTotOps: Integer;
		FDoneOps: Integer;
		FExecStatus: String;
		FDeleteSourceFiles: Boolean;
		procedure SetSourceFiles(srcStrings: TStringList);
		procedure OpenInputStreams;
		procedure GetPages;
		procedure WriteFile;
		procedure CopyPages(fDest: TFileStream; iHeaderSize: Integer);
		procedure RenumberAndCopyPages(fDest: TFileStream; iHeaderSize: Integer);
		procedure UpdateProgress;
	protected
		{ Protected declarations }
	public
		{ Public declarations }
		constructor Create(AOwner: TComponent); override;
		destructor Destroy; override;
		function Execute: Integer;
		property TotOps: Integer read FTotOps;
		property DoneOps: Integer read FDoneOps;
		property ExecStatus: String read FExecStatus;
	published
		{ Published declarations }
		property SourceFiles: TStringList read FSourceFiles write SetSourceFiles;
		property DestinationFile: TFileName read FDestinationFile write FDestinationFile;
		property RenumberPages: Boolean read FRenumberPages write FRenumberPages;
		property OnProgress: TNotifyEvent read FOnProgress write FOnProgress;
		property DeleteSourceFiles: Boolean read FDeleteSourceFiles write FDeleteSourceFiles;
	end;

function QuickSearch(sPattern,sText: String; iFrom: Integer): Integer;
function SearchBack(sPattern,sText: String; iFrom: Integer): Integer;

procedure Register;

implementation

procedure Register;
begin
	RegisterComponents('RBuilder', [TkuRAFMerge]);
end;

constructor TkuRAFMerge.Create(AOwner: TComponent);
	begin
	inherited Create(AOwner);
	FSourceFiles:=TStringList.Create;
	FInputStreams:=TObjectList.Create(True);
	FPages:=TPageList.Create(True);
	FRenumberPages:=False;
	FDeleteSourceFiles:=False;
	end;

destructor TkuRAFMerge.Destroy;
	begin
	FSourceFiles.Free;
	FInputStreams.Free;
	FPages.Free;
	inherited Destroy;
	end;

procedure TkuRAFMerge.SetSourceFiles(srcStrings: TStringList);
	begin
	FSourceFiles.Assign(srcStrings);
	end;

function TkuRAFMerge.Execute: Integer;
	var
		ii: Integer;
	begin
	FTotOps:=90+FSourceFiles.Count;
	FDoneOps:=1;
	Result:=0;

	Try

	//open FSourceFiles, populate FInputStreams and get FTotBytes,FTotPages
	FExecStatus:='Reading file headers.';
	UpdateProgress;
	OpenInputStreams;

	FTotOps:=1+(2*FSourceFiles.Count)+FTotPages;

	//populate FPages using header info in SourceFiles and set FTotOps
	FExecStatus:='Formatting output file.';
	UpdateProgress;
	GetPages;

	//write the output file
	FExecStatus:='Writing output file.';
	UpdateProgress;
	WriteFile;

	//delete source files
	If FDeleteSourceFiles Then
		begin
		//free input streams which will close file handles
		FInputStreams.Clear;
		For ii:=0 to FSourceFiles.Count-1 do
			DeleteFile(FSourceFiles[ii]);
		end;

	//return number of pages processed
	FExecStatus:='Done.';
	Result:=FTotPages;

	Finally
		FPages.Clear;
		FInputStreams.Clear;
		FDoneOps:=FTotOps;
		UpdateProgress;
		end;

	end;

procedure TkuRAFMerge.OpenInputStreams;
	var
		ii,iHeaderSize: Integer;
		fs: TFileStream;
		fID: Array[1..4] of Char;
	begin
	FInputStreams.Clear;
	FInputStreams.Capacity:=FSourceFiles.Count;
	FTotBytes:=0;
	FTotPages:=0;
	For ii:=0 to FSourceFiles.Count-1 do
		begin
		//open each input file and add size to TotBytes to be processed
		If Trim(FSourceFiles[ii])='' Then Continue;
		fs:=TFileStream.Create(FSourceFiles[ii],fmOpenRead or fmShareDenyWrite);
		FTotBytes:=FTotBytes + fs.Size;

		//read first 4 bytes to get number of pages for FTotPages
		// the first 4 bytes indicates the total size of the header
		//  and each page adds 8 bytes to the header
		//the first 4 bytes after the header should be 'TPF0'
		fs.Read(iHeaderSize,4);
		fs.Seek(iHeaderSize,soFromBeginning);
		fs.Read(fID,4);
		If String(fID)<>'TPF0' Then
		 raise exception.Create(FSourceFiles[ii]+' is not a valid RAF file.');

		//update total pages in output
		FTotPages:=FTotPages + (iHeaderSize-4) div 8;

		//save input stream in streams list
		FInputStreams.Add(fs);

		Inc(FDoneOps);
		UpdateProgress;
		end;
	end;

procedure TkuRAFMerge.GetPages;
	var
		iFile,iPage,iPageCount,iHeaderSize,iTmp: Integer;
		iAbsPage,iDestOffset: Integer;
		rbp: TRBPage;
		fs: TFileStream;
	begin
	FPages.Clear;
	FPages.Capacity:=FTotPages;
	//iAbsPage keeps track of cumulative page number
	iAbsPage:=1;
	//iDestOffset keeps track of cumulative page offset into destination
	//  file specified as bytes from end of header
	iDestOffset:=0;

	// for each file parse header and create TRBPage and add to FPages
	For iFile:=0 to FInputStreams.Count-1 do
		begin
		If FInputStreams[iFile]=nil Then Continue;
		fs:=TFileStream(FInputStreams[iFile]);
		//get page count for file
		fs.Seek(0,soFromBeginning);
		fs.Read(iHeaderSize,4);
		iPageCount:=(iHeaderSize-4) div 8;

		//put an RBPage in FPages for each page in file
		For iPage:=1 to iPageCount do
			begin
			rbp:=TRBPage.Create;
			rbp.FSourceIndex:=iFile;
			rbp.FSourcePageNo:=iPage;

			//get offset of page in file (offset is specified as position relative
			// to end of header
			fs.Seek(4 + (iPage-1)*8,soFromBeginning);
			fs.Read(rbp.FSourceOffset,4);

			//add header size to get position of page in file
			rbp.FSourceOffset:=rbp.FSourceOffset+iHeaderSize;

			//get page size
			If iPage=iPageCount Then
				rbp.FPageSize:=fs.Size - rbp.FSourceOffset
			Else
				begin
				//get offset of next page to determine size of this page
				fs.Seek(4 + iPage*8,soFromBeginning);
				fs.Read(iTmp,4);
				iTmp:=iTmp+iHeaderSize;
				rbp.FPageSize:=iTmp - rbp.FSourceOffset;
				end;
			rbp.FDestPageNo:=iAbsPage;
			rbp.FDestOffset:=iDestOffset;
			FPages.Add(rbp);
			Inc(iAbsPage);
			iDestOffset:=iDestOffset + rbp.FPageSize;
			end;

		Inc(FDoneOps);
		UpdateProgress;
		end;
	end;

procedure TkuRAFMerge.WriteFile;
	var
		fd: TFileStream;
		ii,iHeaderSize,iTmp: Integer;
		rbp: TRBPage;
	begin
	FExecStatus:='Opening output file.';
	UpdateProgress;
	fd:=TFileStream.Create(FDestinationFile,fmCreate);
	Try
		//write file header
		//1st 4 bytes gives total header size
		iHeaderSize:=FTotPages*8 + 4;
		fd.Write(iHeaderSize,4);

		//write 8 byte header for each page
		//  1st 4 is offset of page from end of header
		//  2nd 4 is page number
		//  If RenumberPages is requested then the offsets will have to
		//  be overwritten after the pages are potentially resized to accomodate
		//  more digits in the page number than they had.
		FExecStatus:='Writing header.';
		UpdateProgress;
		For ii:=0 to FPages.Count-1 do
			begin
			rbp:=TRBPage(FPages[ii]);
			iTmp:=rbp.FDestOffset;
			fd.Write(iTmp,4);
			iTmp:=rbp.FDestPageNo;
			fd.Write(iTmp,4);
			end;

		//write each page to file and renumber if requested
		FExecStatus:='Writing pages.';
		UpdateProgress;
		If FRenumberPages Then
			RenumberAndCopyPages(fd,iHeaderSize)
		Else
			CopyPages(fd,iHeaderSize);

	Finally
		fd.Free;
		end;
	end;

procedure TkuRAFMerge.CopyPages(fDest: TFileStream; iHeaderSize: Integer);
	var
		fs: TFileStream;
		ii: Integer;
		rbp: TRBPage;
	begin
	For ii:=0 to FPages.Count-1 do
		begin
		rbp:=TRBPage(FPages[ii]);

		//locate page in input file
		fs:=TFileStream(FInputStreams[rbp.FSourceIndex]);
		fs.Seek(rbp.FSourceOffset,soFromBeginning);

		//copy to output file
		fDest.CopyFrom(fs,rbp.FPageSize);

		Inc(FDoneOps);
		UpdateProgress;
		end;
	end;

procedure TkuRAFMerge.RenumberAndCopyPages(fDest: TFileStream; iHeaderSize: Integer);
	var
		fs: TFileStream;
		ss: TStringStream;
		sPg,sDest: String;
		ii,iBlockStart,iBlockEnd,iVTPAGEPos,iTEXTPos: Integer;
		iTmp,iCurSize, iDestOffset: Integer;
		s,sc,sFldText: String;
		rbp: TRBPage;
	begin
	iDestOffset:=0;
	For ii:=0 to FPages.Count-1 do
		begin
		sPg:='';
		//read in page from file to ss stringstream
		rbp:=TRBPage(FPages[ii]);
		fs:=TFileStream(FInputStreams[rbp.FSourceIndex]);
		fs.Seek(rbp.FSourceOffset,soFromBeginning);
		ss:=TStringStream.Create('');
		ss.CopyFrom(fs,rbp.FPageSize);

		//put page into string variable sPg so we can use QuickSearch
		sPg:=ss.DataString;
		ss.Free;

		//we will rewrite sPg to sDest in blocks determined by the position
		// of any vtPage* variables.  Currently supports the following:
		// vtPageNo, vtPageNoDesc, vtPageSet, vtPageSetDesc, vtCount
		// variable is written in sDest in English even if the input file
		// is in another language
		iBlockStart:=1;
		iBlockEnd:=-1;
		sDest:='';

		//search for vtPage* variables
		iVTPAGEPos:=QuickSearch('VTPAGE',sPg,iBlockStart);
		While iVTPAGEPos>0 do
			begin
			//read from iVTPAGEPos to #0 to get type of page number field
			iTmp:=iVTPAGEPos;
			sc:=sPg[iTmp];
			s:='';
			While sc<>Chr(0) do
				begin
				s:=s+sc;
				Inc(iTmp);
				sc:=sPg[iTmp];
				end;

			//create new value for vtPage* field
			If AnsiSameText(s,'vtPageNo') Then
				sFldText:=IntToStr(rbp.FDestPageNo)
			Else If AnsiSameText(s,'vtPageNoDesc') Then
				sFldText:='Page '+IntToStr(rbp.FDestPageNo)
			Else If AnsiSameText(s,'vtPageSet') Then
				sFldText:=IntToStr(rbp.FDestPageNo)+' of '+IntToStr(FTotPages)
			Else If AnsiSameText(s,'vtPageSetDesc') Then
				sFldText:='Page '+IntToStr(rbp.FDestPageNo)+' of '+IntToStr(FTotPages)
			Else If AnsiSameText(s,'vtPageCount') Then
				sFldText:=IntToStr(FTotPages)
			Else sFldText:=' ';

			//search backward from iVTPAGEPos for Text tag
			//  the very closeset it could be is 5 bytes
			iTEXTPos:=SearchBack('TEXT' + #6,sPg,iVTPAGEPos-5);

			//get the current length of the Text field which is given by
			// 2nd byte after the word Text
			iCurSize:=Ord(sPg[iTEXTPos+5]);

			//copy from iBlockStart thru 1 byte after the word Text
			iBlockEnd:=iTEXTPos+4;
			sDest:=sDest+Copy(sPg,iBlockStart,iBlockEnd-iBlockStart+1);

			//then add in the size byte for the new value of Text which will
			//  be larger than iCurSize if the number of digits in the page
			//  number increases
			sDest:=sDest+Chr(Length(sFldText));

			//write the new value of Text
			sDest:=sDest+sFldText;

			//set the new value for iBlockStart
			iBlockStart:=iBlockEnd+iCurSize+2;

			//search for another vtPage* variable
			iVTPAGEPos:=QuickSearch('VTPAGE',sPg,iVTPAGEPOS+12);
			end;
		//write the rest of the page
		sDest:=sDest+Copy(sPg,iBlockStart,Length(sPg)-iBlockStart+1);

		//write sDest to the output file
		ss:=TStringStream.Create(sDest);
		fDest.CopyFrom(ss,ss.Size);

		//update the page size and offset
		rbp.FPageSize:=ss.Size;
		rbp.FDestOffset:=iDestOffset;
		iDestOffset:=iDestOffset+rbp.FPageSize;
		ss.Free;

		Inc(FDoneOps);
		UpdateProgress;
		end;

	//overwrite page offsets in output file header
	For ii:=0 to FPages.Count-1 do
		begin
		rbp:=TRBPage(FPages[ii]);
		iTmp:=rbp.FDestOffset;
		fDest.Position:=4 + ii*8;
		fDest.Write(iTmp,4);
		end;
	end;

procedure TkuRAFMerge.UpdateProgress;
	begin
	If Assigned(FOnProgress) Then FOnProgress(self);
	end;

function QuickSearch(sPattern,sText: String; iFrom: Integer): Integer;
	var
		qsAlpha: Array[0..26] of Char;
		qsFunc: Array[0..26] of Integer;
		ii,iPatLen,iTextLen: Integer;
		s,sc,sPatUp: String;
	begin
	Result:=-1;

	//determine QuickSearch Preprocessing Function
	// search will be case insensitive
	qsAlpha:='ABCDEFGHIJKLMNOPQRSTUVWXYZ~';

	sPatUp:=UpperCase(sPattern);
	iPatLen:=Length(sPattern);
	iTextLen:=Length(sText);
	SetLength(s,iPatLen);
	SetLength(sc,1);

	For ii:=0 to Length(qsAlpha)-1 do qsFunc[ii]:=iPatLen+1;
	For ii:=0 to iPatLen-1 do
		begin
		s:=UpperCase(sPattern[ii+1]);
		qsFunc[Pos(s,qsAlpha)-1]:=iPatLen-ii;
		end;

	//search for first occurence beginning at iFrom
	ii:=iFrom;
	While ii < iTextLen-iPatLen do
		begin
		s:=UpperCase(Copy(sText,ii,iPatLen));
		If s=sPatUp Then
			begin
			Result:=ii;
			Exit;
			end
		Else
			begin
			sc:=UpperCase(sText[ii+iPatLen]);
			If (sc<'A') OR (sc>'Z') Then sc:='~';
			ii:=ii+qsFunc[Pos(sc,qsAlpha)-1];
			end;
		end;
	end;


function SearchBack(sPattern,sText: String; iFrom: Integer): Integer;
	var
		ii,iPatLen,iTextLen: Integer;
		s,sc,sPatUp: String;
	begin
	//lazy way to search backward since it won't be very far from
	//starting point plus i won't have to figure out how to do
	//quicksort backwards

	Result:=-1;

	sPatUp:=UpperCase(sPattern);
	iPatLen:=Length(sPattern);
	iTextLen:=Length(sText);
	SetLength(s,iPatLen);
	SetLength(sc,1);

	ii:=iFrom;
	While ii>0 do
		begin
		s:=UpperCase(Copy(sText,ii,iPatLen));
		If s=sPatUp Then
			begin
			Result:=ii;
			Exit;
			end
		Else
			Dec(ii);
		end;
	end;
end.
